@@ -2,7 +2,7 @@
  *                                                                         *
  *   libgig - C++ cross-platform Gigasampler format file access library    *
  *                                                                         *
- *   Copyright (C) 2003-2019 by Christian Schoenebeck                      *
+ *   Copyright (C) 2003-2021 by Christian Schoenebeck                      *
  *                              <cuse@users.sourceforge.net>               *
  *                                                                         *
  *   This library is free software; you can redistribute it and/or modify  *
@@ -44,6 +44,12 @@
 # endif
 #endif
 
+#ifdef __GNUC__
+# define LIBGIG_DEPRECATED_API(msg) __attribute__ ((deprecated(msg)))
+#else
+# define LIBGIG_DEPRECATED_API(msg)
+#endif
+
 #include <string>
 #include <list>
 #include <map>
@@ -51,6 +57,8 @@
 #include <vector>
 #include <iostream>
 #include <stdarg.h>
+#include <thread>
+#include <mutex>
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -238,8 +246,8 @@ namespace RIFF {
             List*          GetParent() const  { return pParent; }       ///< Returns pointer to the chunk's parent list chunk.
             file_offset_t  GetSize() const { return ullCurrentChunkSize; } ///< Chunk size in bytes (without header, thus the chunk data body)
             file_offset_t  GetNewSize() const { return ullNewChunkSize; } ///< New chunk size if it was modified with Resize(), otherwise value returned will be equal to GetSize().
-            file_offset_t  GetPos() const { return ullPos; }            ///< Position within the chunk data body (starting with 0).
-            file_offset_t  GetFilePos() const { return ullStartPos + ullPos; } ///< Current, actual offset in file of current chunk data body read/write position.
+            file_offset_t  GetPos() const;
+            file_offset_t  GetFilePos() const;
             file_offset_t  SetPos(file_offset_t Where, stream_whence_t Whence = stream_start);
             file_offset_t  RemainingBytes() const;
             stream_state_t GetState() const;
@@ -275,9 +283,13 @@ namespace RIFF {
             List*         pParent;
             File*         pFile;
             file_offset_t ullStartPos;  /* actual position in file where chunk data (without header) starts */
-            file_offset_t ullPos;       /* # of bytes from ulStartPos */
             uint8_t*      pChunkData;
             file_offset_t ullChunkDataSize;
+            struct ChunkPos { /* current read/write position as # of bytes from ulStartPos */
+                file_offset_t ullPos; /* used if File::IsIOPerThread() is false (default) */
+                mutable std::map<std::thread::id,file_offset_t> byThread; /* used if File::IsIOPerThread() is true */
+                mutable std::mutex mutex; /* for protecting concurrent changes on byThread map itself */
+            } chunkPos;
 
             Chunk(File* pFile);
             Chunk(File* pFile, List* pParent, uint32_t uiChunkID, file_offset_t ullBodySize);
@@ -298,6 +310,8 @@ namespace RIFF {
             virtual void __resetPos(); ///< Sets Chunk's read/write position to zero.
 
             friend class List;
+        private:
+            file_offset_t& GetPosUnsafeRef();
     };
 
     /** @brief RIFF List Chunk
@@ -310,12 +324,14 @@ namespace RIFF {
             List(File* pFile, file_offset_t StartPos, List* Parent);
             String       GetListTypeString() const;
             uint32_t     GetListType() const { return ListType; } ///< Returns unsigned integer representation of the list's ID
+            Chunk*       GetSubChunkAt(size_t pos);
             Chunk*       GetSubChunk(uint32_t ChunkID);
+            List*        GetSubListAt(size_t pos);
             List*        GetSubList(uint32_t ListType);
-            Chunk*       GetFirstSubChunk();
-            Chunk*       GetNextSubChunk();
-            List*        GetFirstSubList();
-            List*        GetNextSubList();
+            Chunk*       GetFirstSubChunk() LIBGIG_DEPRECATED_API("Use GetSubChunkAt() instead.");
+            Chunk*       GetNextSubChunk() LIBGIG_DEPRECATED_API("Use GetSubChunkAt() instead.");
+            List*        GetFirstSubList() LIBGIG_DEPRECATED_API("Use GetSubListAt() instead.");
+            List*        GetNextSubList() LIBGIG_DEPRECATED_API("Use GetSubListAt() instead.");
             size_t       CountSubChunks();
             size_t       CountSubChunks(uint32_t ChunkID);
             size_t       CountSubLists();
@@ -328,7 +344,7 @@ namespace RIFF {
             virtual ~List();
         protected:
             typedef std::map<uint32_t, RIFF::Chunk*>  ChunkMap;
-            typedef std::list<Chunk*>                 ChunkList;
+            typedef std::vector<Chunk*>               ChunkList;
             typedef std::set<Chunk*>                  ChunkSet;
 
             uint32_t   ListType;
@@ -357,6 +373,18 @@ namespace RIFF {
      */
     class File : public List {
         public:
+            /**
+             * OS dependent type serving as file handle / descriptor for OS
+             * dependent file I/O operations.
+             */
+            #if POSIX
+            typedef int Handle;
+            #elif defined(WIN32)
+            typedef HANDLE Handle;
+            #else
+            typedef FILE* Handle;
+            #endif
+
             File(uint32_t FileType);
             File(const String& path);
             File(const String& path, uint32_t FileType, endian_t Endian, layout_t layout, offset_size_t fileOffsetSize = offset_size_auto);
@@ -365,6 +393,7 @@ namespace RIFF {
             void SetByteOrder(endian_t Endian);
             String GetFileName() const;
             void SetFileName(const String& path);
+            Handle FileHandle() const;
             bool IsNew() const;
             layout_t GetLayout() const;
             file_offset_t GetCurrentFileSize() const;
@@ -372,21 +401,23 @@ namespace RIFF {
             file_offset_t GetRequiredFileSize(offset_size_t fileOffsetSize);
             int GetFileOffsetSize() const;
             int GetRequiredFileOffsetSize();
+            bool IsIOPerThread() const;
+            void SetIOPerThread(bool enable);
 
             virtual void Save(progress_t* pProgress = NULL);
             virtual void Save(const String& path, progress_t* pProgress = NULL);
             virtual ~File();
         protected:
-            #if POSIX
-            int    hFileRead;  ///< handle / descriptor for reading from file
-            int    hFileWrite; ///< handle / descriptor for writing to (some) file
-            #elif defined(WIN32)
-            HANDLE hFileRead;  ///< handle / descriptor for reading from file
-            HANDLE hFileWrite; ///< handle / descriptor for writing to (some) file
-            #else
-            FILE*  hFileRead;  ///< handle / descriptor for reading from file
-            FILE*  hFileWrite; ///< handle / descriptor for writing to (some) file
-            #endif // POSIX
+            struct HandlePair {
+                Handle hRead;  ///< handle / descriptor for reading from file
+                Handle hWrite; ///< handle / descriptor for writing to (some) file
+                stream_mode_t Mode; ///< State of handles (ro, rw, closed).
+            };
+            struct IO : HandlePair {
+                bool isPerThread; ///< Whether SetIOPerThread(true) was called.
+                mutable std::map<std::thread::id,HandlePair> byThread; ///< Optional: Individual file handle pair for each thread separately.
+                mutable std::mutex mutex; ///< For protecting concurrent changes on the @c byThread map itself and replacing handles.
+            } io;
             String Filename;
             bool   bEndianNative;
             bool   bIsNewFile;
@@ -394,11 +425,12 @@ namespace RIFF {
             offset_size_t FileOffsetPreference;
             int FileOffsetSize; ///< Size of file offsets (in bytes) when this file was opened (or saved the last time).
 
+            Handle FileWriteHandle() const;
+            HandlePair FileHandlePair() const;
+
             friend class Chunk;
             friend class List;
         private:
-            stream_mode_t  Mode;
-
             void __openExistingFile(const String& path, uint32_t* FileType = NULL);
             void ResizeFile(file_offset_t ullNewSize);
             #if POSIX
@@ -410,6 +442,8 @@ namespace RIFF {
             #endif
             int FileOffsetSizeFor(file_offset_t fileSize) const;
             void Cleanup();
+            HandlePair& FileHandlePairUnsafeRef();
+            bool SetModeInternal(stream_mode_t NewMode, bool* pResetPos);
     };
 
     /**
